@file:Suppress("unused")

package abstractmath

fun <El, T : CayleyDicksonConstruction<El, T>> Re(complex: T) = complex.re

fun <El, T : CayleyDicksonConstruction<El, T>> Im(complex: T) = complex.im

/**
 * Specifies the Cayley-Dickson Construction over El field
 *
 * Denote C-D C as a + bI, where a, b are elements of El field, I is imaginary unit: I^2 = γ, but I
 * is not element of El field
 *
 * a is called as "real part", b is called as "imaginary part"
 *
 * There are three kind of C-D C:
 *
 * With γ = -1 (Elliptic C-D C, or Complex)
 *
 * With γ = 0 (Parabolic C-D C, or Dual)
 *
 * With γ = 1 (Hyperbolic C-D C, or Split)
 *
 * Then:
 *
 * (a + bI) + (c + dI) = (a + c) + (b + d)I
 *
 * -(a + bI) = (-a) + (-b)I
 *
 * (a + bI) x (c + dI) = (a c + γ d* b) + (d a + b c*)I
 *
 * (a + bI)^(-1) = *(a + bI) / ||a + bI|| = (a / ||a + bI||) + (-b / ||a + bI||)I
 *
 * (a + bI)* = a* + (-b)I
 *
 * ||a + bI|| = a x a + b x b
 *
 * zero is (0 + 0I)
 *
 * one is (1 + 0I)
 */
sealed class CayleyDicksonConstruction<El : Field<El>, Self : CayleyDicksonConstruction<El, Self>>
(val re: El, val im: El) : Field<Self> {

    abstract fun gamma(): Int
    protected abstract fun construct(real: El, imaginary: El): Self

    override operator fun plus(another: Self) = construct(re + another.re, im + another.im)

    override operator fun unaryMinus() = construct(-re, -im)

    override fun inverse() = construct(re.involution() / norm().re, -im / norm().re)
    //norm should be real-only, but returned CayleyDicksonConstruction

    override operator fun minus(another: Self) = construct(re - another.re, im - another.im)

    override fun involution() = construct(re.involution(), -im)

    operator fun plus(x: El) = construct(re + x, im)
    operator fun minus(x: El) = construct(re - x, im)

    override fun zero() = construct(re.zero(), im.zero()) // Just the same zeros ..-)
    override fun one() = construct(re.one(), im.zero())

    override fun toString() = "$re + I($im)"
}

class EllipticCayleyDicksonConstruction<El : Field<El>>(real: El, imaginary: El) :
        CayleyDicksonConstruction<El, EllipticCayleyDicksonConstruction<El>>(real, imaginary) {

    override fun gamma() = -1
    override fun construct(real: El, imaginary: El) = EllipticCayleyDicksonConstruction(real, imaginary)

    override operator fun times(another: EllipticCayleyDicksonConstruction<El>) =
            EllipticCayleyDicksonConstruction(
                    re * another.re - another.im.involution() * im,
                    another.im * re + im * another.re.involution()
            )
}


class ParabolicCayleyDicksonConstruction<El : Field<El>>(real: El, imaginary: El) :
        CayleyDicksonConstruction<El, ParabolicCayleyDicksonConstruction<El>>(real, imaginary) {

    override fun gamma() = 0
    override fun construct(real: El, imaginary: El) = ParabolicCayleyDicksonConstruction(real, imaginary)

    override operator fun times(another: ParabolicCayleyDicksonConstruction<El>) =
            ParabolicCayleyDicksonConstruction(
                    re * another.re,
                    another.im * re + im * another.re.involution()
            )
}

class HyperbolicCayleyDicksonConstruction<El : Field<El>>(real: El, imaginary: El) :
        CayleyDicksonConstruction<El, HyperbolicCayleyDicksonConstruction<El>>(real, imaginary) {

    override fun gamma() = 1
    override fun construct(real: El, imaginary: El) = HyperbolicCayleyDicksonConstruction(real, imaginary)

    override operator fun times(another: HyperbolicCayleyDicksonConstruction<El>) =
            HyperbolicCayleyDicksonConstruction(
                    re * another.re + another.im.involution() * im,
                    another.im * re + im * another.re.involution()
            )
}

