﻿package abstractmath

import javax.naming.OperationNotSupportedException
import kotlin.math.min

class EuclideanVector<El : Field<El>> {
    val list: List<El>
    val fieldzero: El

    constructor(list: List<El>) {
        if (list.isEmpty()) throw OperationNotSupportedException("Zero-dimensional vectors are not supported yet")
        this.list = list
        fieldzero = list[0].zero()
    }

    constructor(vararg elements: El) : this(elements.toList())

    @Suppress("SpellCheckingInspection")
    operator fun plus(another: EuclideanVector<El>): EuclideanVector<El> {
        val newlist = MutableList(maxOf(list.size, another.list.size)) { fieldzero }
        for (i in list.indices) newlist[i] = list[i]
        for (i in list.indices) newlist[i] += another.list[i]
        return EuclideanVector(newlist)
    }

    operator fun unaryMinus() = EuclideanVector(list.map { -it })

    operator fun minus(another: EuclideanVector<El>) = this + (-another)

    operator fun times(another: El) = EuclideanVector(list.map { it * another })

    operator fun div(another: El) = EuclideanVector(list.map { it / another })

    operator fun get(index: Int) = list[index]

    operator fun get(index: Int, default: El) = if (index in list.indices) list[index] else default

    fun scalarProduct(another: EuclideanVector<El>): El {
        var result = fieldzero
        for (i in 0 until min(list.size, another.list.size)) result += list[i] * another.list[i]
        return result
    }

    /**
     * | Ex Ey Ez |
     *
     * | X₁ Y₁ Z₁ |
     *
     * | X₂ Y₂ Z₂ |
     *
     * Ex (Y₁Z₂ - Z₁Y₂) - Ey (X₁Z₂ - Z₁X₂) + Ez (X₁Y₂ - Y₁X₂)
     */
    fun vectorProduct(another: EuclideanVector<El>): EuclideanVector<El> {
        if (list.size > 3 || another.list.size > 3) TODO("Multiple argument vector product for more than third dimension vectors are not supported yet")

        val x1 = this[0, fieldzero]
        val y1 = this[1, fieldzero]
        val z1 = this[2, fieldzero]
        val x2 = another[0, fieldzero]
        val y2 = another[1, fieldzero]
        val z2 = another[2, fieldzero]
        return EuclideanVector(y1 * z2 - z1 * y2, z1 * x2 - x1 * z2, x1 * y2 - y1 * x2)
    }

    infix fun x(another: EuclideanVector<El>) = vectorProduct(another)

    fun norm() = scalarProduct(this)

    override fun toString() = list.toString()
}


operator fun <El : Field<El>> El.times(another: EuclideanVector<El>) =
        EuclideanVector(another.list.map { this * it })
//Order is important! There are algebras with non-commutative multiplication


