package abstractmath

/**
 * Specifies the field with involution
 *
 * Denote:
 *
 * a + b - addition
 *
 * a - b - subtraction
 *
 * -a - unary subtraction
 *
 * +a - unary addition
 *
 * 0 - zero
 *
 * 1 - one
 * .
 * a x b - multiplication
 *
 * a^(-1) - inversion (exists for any a != 0)
 *
 * a / b - division
 *
 * a* - involution
 *
 * ||a|| - norm
 *
 *
 * The following properties must be satisfied:
 *
 * a - b = a + (-b)
 *
 * -(-a) = a
 *
 * a / b = a x (b^(-1))
 *
 * +a = a
 *
 * (a*)* = a
 *
 * ||a|| = a x (a*)
 *
 * a + b = b + a
 *
 * a + (b + c) = (a + b) + c
 *
 * a + 0 = a
 *
 * -0 = 0
 *
 * a - a = 0
 *
 * a x (b + c) = a x b + a x c
 *
 * (a + b) x c = a x c + b x c
 *
 * a / a = 1 //For all a != 0
 *
 * 1^(-1) = 1
 *
 * ((a^(-1))^(-1) = a //For all a != 0
 *
 * (-a) x b = a x (-b) = - (a x b)
 *
 * a x 1 = 1 x a = a
 *
 * Note that multiplication may be not commutative and not associative!
 *
 * <Self> must be self class, to avoid ClassCastException during (this as Self) casting
 */ //FIXME add some properties? I'm bad in field theory...

interface Field<Self : Field<Self>> {
    operator fun plus(another: Self): Self
    operator fun unaryMinus(): Self
    operator fun times(another: Self): Self
    fun inverse(): Self

    @Suppress("UNCHECKED_CAST")
    operator fun unaryPlus(): Self = this as Self

    operator fun minus(another: Self) = this + (-another)
    operator fun div(another: Self) = this * another.inverse()

    fun involution(): Self
    fun norm() = this * involution()

    fun zero(): Self
    fun one(): Self
}

