package graphics

import abstractmath.*
import abstractmath.Real.Companion.R0
import abstractmath.Real.Companion.R1
import graphics.figures.VectorBase
import graphics.figures.VectorGraphicsProvider
import java.awt.Color
import java.awt.Graphics
import kotlin.math.exp
import kotlin.math.sqrt

data class Vertex(
        var label: String,
        var coordID: Int,
        var showLabel: Boolean = true,
        var showPoint: Boolean = true,
        var color: Color = Color.white,
        var pointSize: Int = 5
) : VectorGraphicsProvider {
    override fun paint(g: Graphics, vb: VectorBase, scale: Double) {
        if (!showPoint && !showLabel) return

        g.color = color
        val vector = vb.getVector(coordID)
        val perspectiveCoeff = perspectiveCoeff(vector)
        val x = (vector[0].toDouble() * perspectiveCoeff / scale).toInt()
        val y = (vector[1].toDouble() * perspectiveCoeff / scale).toInt()

        var labelShift = 3
        if (showPoint) {
            g.fillOval(x - pointSize, y - pointSize, pointSize * 2, pointSize * 2)
            labelShift += pointSize
        }
        if (showLabel) g.drawString(label, x + labelShift, y - labelShift)
    }
}

data class Edge(val aID: Int, val bID: Int, var color: Color, var show: Boolean = true) : VectorGraphicsProvider {
    override fun paint(g: Graphics, vb: VectorBase, scale: Double) {
        if (!show) return

        val a = vb.getVector(aID)
        val perspectiveCoeffA = perspectiveCoeff(a)
        val xa = (a[0].toDouble() * perspectiveCoeffA / scale).toInt()
        val ya = (a[1].toDouble() * perspectiveCoeffA / scale).toInt()

        val b = vb.getVector(bID)
        val perspectiveCoeffB = perspectiveCoeff(b)
        val xb = (b[0].toDouble() * perspectiveCoeffB / scale).toInt()
        val yb = (b[1].toDouble() * perspectiveCoeffB / scale).toInt()

        g.color = color
        g.drawLine(xa, ya, xb, yb)
    }
}

data class ConvexFace(
        val points: List<Int>,
        var color: Color = Color.green,
        var shown: Boolean = false
) : VectorGraphicsProvider {
    constructor(vararg points: Int, color: Color) :
            this(points.toList(), color)

    override fun paint(g: Graphics, vb: VectorBase, scale: Double) {
        shown = false
//        if (!show) return

        val v1 = vb.getVector(points[0])
        val v2 = vb.getVector(points[1])
        val v3 = vb.getVector(points[2])
        val normal = ((v2 - v1) x (v3 - v2))

        if (normal.scalarProduct(FRONT) <= R0) return

        val xArray = IntArray(points.size)
        val yArray = IntArray(points.size)

        for (i in points.indices) {
            val vertex = vb.getVector(points[i])
            val perspectiveCoeff = perspectiveCoeff(vertex)
            xArray[i] = (vertex[0].toDouble() * perspectiveCoeff / scale).toInt()
            yArray[i] = (vertex[1].toDouble() * perspectiveCoeff / scale).toInt()
        }


        g.color = enlight(normal, color)
        g.fillPolygon(xArray, yArray, points.size)
        shown = true
    }

    private fun enlight(normal: EVector, color: Color): Color {
        val angleCos = -normal.scalarProduct(LIGHT_VECTOR) / normal.magnitude().R / LIGHT_VECTOR.magnitude().R
        val lightCoeff = (angleCos.toDouble() + 1)/2
        return colorOf(color.red * lightCoeff, color.green * lightCoeff, color.blue * lightCoeff)
    }

    private fun colorOf(r: Number, g: Number, b: Number): Color {
        var rnew = r.toInt()
        var gnew = g.toInt()
        var bnew = b.toInt()

        if (rnew < 0) rnew = 0
        if (rnew > 255) rnew = 255
        if (gnew < 0) gnew = 0
        if (gnew > 255) gnew = 255
        if (bnew < 0) bnew = 0
        if (bnew > 255) bnew = 255
        return Color(rnew, gnew, bnew)
    }

    companion object {
        val FRONT = EVector(R0, R0, R1)
    }

}

fun perspectiveCoeff(vertex: EVector) = 1 //Math.pow(1.02, -vertex[2].toDouble())

fun EVector.magnitude() = sqrt(norm().toDouble())
