package graphics.figures

import abstractmath.R
import abstractmath.Real
import abstractmath.Real.Companion.R0
import graphics.ConvexFace
import graphics.Edge
import graphics.Vertex
import graphics.magnitude
import java.awt.Color

open class Figure : FigureGraphicsProvider() {
    protected val Real.pm: List<Real> get() = listOf(this, -this)
    protected val Number.pm: List<Real> get() = listOf(this.R, -this.R)

    protected fun allCombinations(xes: List<Real>, ys: List<Real>, zs: List<Real>) {
        for (x in xes) for (y in ys) for (z in zs) vb.addVector(x, y, z)
    }

    protected fun circlePermutations(xes: List<Real>, ys: List<Real>, zs: List<Real>) {
        allCombinations(xes, ys, zs)
        allCombinations(zs, xes, ys)
        allCombinations(ys, zs, xes)
    }

    protected fun allPermutations(xes: List<Real>, ys: List<Real>, zs: List<Real>) {
        allCombinations(xes, ys, zs)
        allCombinations(xes, zs, ys)
        allCombinations(ys, xes, zs)
        allCombinations(ys, zs, xes)
        allCombinations(zs, xes, ys)
        allCombinations(zs, ys, xes)
    }

    protected fun verticesOfAllVectors(
            labels: List<String>,
            showLabel: Boolean = true,
            showPoint: Boolean = true,
            color: Color = Color.white,
            pointSize: Int = 5
    ) {
        for (i in 0 until vb.size) {
            vertices.add(Vertex(labels[i], i, showLabel, showPoint, color, pointSize))
        }
    }

    protected fun verticesOfAllVectors(
            showLabel: Boolean = true,
            showPoint: Boolean = true,
            color: Color = Color.white,
            pointSize: Int = 5
    ) {
        for (i in 0 until vb.size) {
            vertices.add(Vertex("", i, showLabel, showPoint, color, pointSize))
        }
    }
    protected fun edgesOfLength(length: Double, color: Color, show: Boolean = true) {
        val lengthRange = (length * 0.95)..(length * 1.05)
        for (i in 0 until vb.size) for (j in (i + 1) until vb.size) {
            if ((vb.getVector(i) - vb.getVector(j)).magnitude() in lengthRange) {
                edges.add(Edge(i, j, color, show))
            }
        }
    }

    @Deprecated("Very long method: about O(vb.size ^ verticesInFace)")
    protected fun facesOfVertices(verticesInFace: Int, color: Color) {
        assert(verticesInFace > 2) { "It has no sense to create faces of less than 3 vertices" }

        forAllCombinations(vb.size, verticesInFace) { createFaceIfNeed(it, color) }
    }

    private fun createFaceIfNeed(vertices: List<Int>, color: Color) {
        //Each vertex must have two edges with others of list
        for (vertex in vertices) if (vertices.filter { doesEdgeExist(it, vertex) }.size != 2) return

        val order = mutableListOf<Int>()
        order.add(vertices[0])

        do {
            var changed = false
            for (v in vertices) {
                if (v !in order && doesEdgeExist(v, order.last())) {
                    order.add(v)
                    changed = true
                }
            }
        } while (changed)

        if (order.size != vertices.size) return

        val v1 = vb.getVector(order[0])
        val v2 = vb.getVector(order[1])
        val v3 = vb.getVector(order[2])
        val normal = ((v2 - v1) x (v3 - v2))

        if (normal.scalarProduct(v2) <= R0) { //I'm not sure this works correctly
            order.reverse()
        }

        faces.add(ConvexFace(order, color))
    }

    private fun doesEdgeExist(a: Int, b: Int) = edges.any { (it.aID == a && it.bID == b) || (it.aID == b && it.bID == a) }

    //For example, when sample = 3, length = 2, combinations are (0, 1), (0, 2), (1, 2)
    private fun forAllCombinations(sample: Int, length: Int, prefix: List<Int> = listOf(), begin: Int = 0, function: (List<Int>) -> Unit) {
        if (length == 0) return function(prefix)

        for (i in begin..(sample - length)) {
            forAllCombinations(sample, length - 1, prefix + i, i + 1, function)
        }
    }
}