@file:Suppress("unused")

package graphics.figures

import abstractmath.*
import abstractmath.Real.Companion.R1
import abstractmath.Real.Companion.R0
import graphics.*
import java.awt.Color
import java.awt.Graphics
import kotlin.math.*

interface VectorGraphicsProvider {
    fun paint(g: Graphics, vb: VectorBase, scale: Double)
}

open class FigureGraphicsProvider : GraphicsProvider {
    protected val vertices = mutableListOf<Vertex>()
    protected val edges = mutableListOf<Edge>()
    protected val faces = mutableListOf<ConvexFace>()
    protected val vb = VectorBase()

    override fun paint(g: Graphics, width: Int, height: Int) {
        if (vb.isEmpty()) return
        g.color = Color.black
        g.fillRect(0, 0, width, height)
        g.translate(width / 2, height / 2)

        val maxNorm = vb.list().maxByOrNull { it.norm().toDouble() }!!
        val scale = 2 * 2 * sqrt(maxNorm.norm().toDouble()) / width

        if (showCarcassOnly) {
            for (e in edges) e.paint(g, vb, scale)
            for (v in vertices) v.paint(g, vb, scale)
        } else {
            for (f in faces) f.paint(g, vb, scale)
            for (e in edges) {
                if (faces.any { it.shown && e.aID in it.points && e.bID in it.points })
                    e.paint(g, vb, scale)
            }
            for (v in vertices) {
                if (faces.any { it.shown && v.coordID in it.points })
                    v.paint(g, vb, scale)
            }
        }
    }

    fun rotate(angle: Number, axis: EVector) = vb.rotate(angle, axis)
}

class VectorBase : Iterable<EVector> {
    private val list = mutableListOf<EVector>()
    private var lastId = 0
    private var rotation = quaternionOf(R1, R0, R0, R0)
    private var rotationInverse = quaternionOf(R1, R0, R0, R0)

    fun list() = list as List<EVector>

    fun isEmpty() = list.isEmpty()
    fun isNotEmpty() = list.isNotEmpty()

    val size: Int
        get() = list.size


    fun addVector(vararg coords: Real): Int {
        list.add(EVector(coords.toList()))
        return lastId++
    }

    fun addVector(vararg coords: Number): Int {
        list.add(EVector(coords.map { it.R }))
        return lastId++
    }

    fun getVector(id: Int): EVector {
        if (id !in list.indices) throw IllegalAccessException("No such vector found for id$id")

        synchronized(rotation) {
            return (rotation * (R0 + list[id]) * rotationInverse).vector()
        }
    }

    companion object {
        const val normalizingPeriod = 25
    }

    var timesRotated = 0
    fun rotate(angle: Number, axis: EVector) {
        val normalized = axis / axis.magnitude()
        val rotor = cos(angle.toDouble() / 2).R + sin(angle.toDouble() / 2) * normalized

        synchronized(rotation) {
            rotation *= rotor

            if (++timesRotated == normalizingPeriod) {
                timesRotated = 0
                rotation /= quaternionOf(sqrt(rotation.norm().re.re.toDouble()).R, R0, R0, R0)
            }

            rotationInverse = rotation.inverse()
        }
    }

    override fun iterator() = List(list.size) { getVector(it) }.iterator()
}

